#include <sys/types.h>
#include <unistd.h>
#include <stdint.h>

#include "uECC.h"

typedef uint32_t uECC_word_t;
typedef uint64_t uECC_dword_t;
typedef unsigned wordcount_t;
typedef int swordcount_t;
typedef int bitcount_t;
typedef int cmpresult_t;
typedef struct EccPoint
{
    uECC_word_t x[8];
    uECC_word_t y[8];
} EccPoint;
static uECC_word_t curve_p[8] = {0xFFFFFC2F, 0xFFFFFFFE, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF};
static uECC_word_t curve_b[8] = {0x00000007, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000, 0x00000000};
static EccPoint curve_G = { {0x16F81798, 0x59F2815B, 0x2DCE28D9, 0x029BFCDB, 0xCE870B07, 0x55A06295, 0xF9DCBBAC, 0x79BE667E}, {0xFB10D4B8, 0x9C47D08F, 0xA6855419, 0xFD17B448, 0x0E1108A8, 0x5DA4FBFC, 0x26A3C465, 0x483ADA77}};
static uECC_word_t curve_n[8] = {0xD0364141, 0xBFD25E8C, 0xAF48A03B, 0xBAAEDCE6, 0xFFFFFFFE, 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFF};
static void vli_clear(uECC_word_t *p_vli);
static uECC_word_t vli_isZero(const uECC_word_t *p_vli);
static uECC_word_t vli_testBit(const uECC_word_t *p_vli, bitcount_t p_bit);
static bitcount_t vli_numBits(const uECC_word_t *p_vli, wordcount_t p_maxWords);
static void vli_set(uECC_word_t *p_dest, const uECC_word_t *p_src);
static cmpresult_t vli_cmp(uECC_word_t *p_left, uECC_word_t *p_right);
static cmpresult_t vli_equal(uECC_word_t *p_left, uECC_word_t *p_right);
static void vli_rshift1(uECC_word_t *p_vli);
static uECC_word_t vli_add(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right);
static uECC_word_t vli_sub(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right);
static void vli_mult(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right);
static void vli_modAdd(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right, uECC_word_t *p_mod);
static void vli_modSub(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right, uECC_word_t *p_mod);
static void vli_mmod_fast(uECC_word_t * p_result, uECC_word_t * p_product);
static void vli_modMult_fast(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right);
static void vli_modInv(uECC_word_t *p_result, uECC_word_t *p_input, uECC_word_t *p_mod);
static void vli_square(uECC_word_t *p_result, uECC_word_t *p_left);
static void vli_modSquare_fast(uECC_word_t *p_result, uECC_word_t *p_left);

static int default_RNG(uint8_t *p_dest, unsigned p_size)
{
    int l_fd = open("/dev/urandom", 00 | 02000000);
    if(l_fd == -1)
    {
        l_fd = open("/dev/random", 00 | 02000000);
        if(l_fd == -1)
        {
            return 0;
        }
    }
    char *l_ptr = (char *)p_dest;
    size_t l_left = p_size;
    while(l_left > 0)
    {
        int l_read = read(l_fd, l_ptr, l_left);
        if(l_read <= 0)
        {
            close(l_fd);
            return 0;
        }
        l_left -= l_read;
        l_ptr += l_read;
    }
    close(l_fd);
    return 1;
}
static uECC_RNG_Function g_rng = &default_RNG;
void uECC_set_rng(uECC_RNG_Function p_rng)
{
    g_rng = p_rng;
}
static void vli_clear(uECC_word_t *p_vli)
{
    wordcount_t i;
    for(i = 0; i < 8; ++i)
    {
        p_vli[i] = 0;
    }
}
static uECC_word_t vli_isZero(const uECC_word_t *p_vli)
{
    wordcount_t i;
    for(i = 0; i < 8; ++i)
    {
        if(p_vli[i])
        {
            return 0;
        }
    }
    return 1;
}
static uECC_word_t vli_testBit(const uECC_word_t *p_vli, bitcount_t p_bit)
{
    return (p_vli[p_bit >> 5] & ((uECC_word_t)1 << (p_bit & 0x01F)));
}
static wordcount_t vli_numDigits(const uECC_word_t *p_vli, wordcount_t p_maxWords)
{
    swordcount_t i;
    for(i = p_maxWords-1; i >= 0 && p_vli[i] == 0; --i)
    {
    }
    return (i + 1);
}
static bitcount_t vli_numBits(const uECC_word_t *p_vli, wordcount_t p_maxWords)
{
    uECC_word_t i;
    uECC_word_t l_digit;
    wordcount_t l_numDigits = vli_numDigits(p_vli, p_maxWords);
    if(l_numDigits == 0)
    {
        return 0;
    }
    l_digit = p_vli[l_numDigits - 1];
    for(i = 0; l_digit; ++i)
    {
        l_digit >>= 1;
    }
    return (((bitcount_t)(l_numDigits - 1) << 5) + i);
}
static void vli_set(uECC_word_t *p_dest, const uECC_word_t *p_src)
{
    wordcount_t i;
    for(i=0; i<8; ++i)
    {
        p_dest[i] = p_src[i];
    }
}
static cmpresult_t vli_cmp(uECC_word_t *p_left, uECC_word_t *p_right)
{
    swordcount_t i;
    for(i = 8 -1; i >= 0; --i)
    {
        if(p_left[i] > p_right[i])
        {
            return 1;
        }
        else if(p_left[i] < p_right[i])
        {
            return -1;
        }
    }
    return 0;
}
static cmpresult_t vli_equal(uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_result = 0;
    swordcount_t i;
    for(i = 8 -1; i >= 0; --i)
    {
        l_result |= (p_left[i] ^ p_right[i]);
    }
    return (l_result == 0);
}
static void vli_rshift1(uECC_word_t *p_vli)
{
    uECC_word_t *l_end = p_vli;
    uECC_word_t l_carry = 0;
    p_vli += 8;
    while(p_vli-- > l_end)
    {
        uECC_word_t l_temp = *p_vli;
        *p_vli = (l_temp >> 1) | l_carry;
        l_carry = l_temp << (32 - 1);
    }
}
static uECC_word_t vli_add(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_carry = 0;
    wordcount_t i;
    for(i = 0; i < 8; ++i)
    {
        uECC_word_t l_sum = p_left[i] + p_right[i] + l_carry;
        if(l_sum != p_left[i])
        {
            l_carry = (l_sum < p_left[i]);
        }
        p_result[i] = l_sum;
    }
    return l_carry;
}
static uECC_word_t vli_sub(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_borrow = 0;
    wordcount_t i;
    for(i = 0; i < 8; ++i)
    {
        uECC_word_t l_diff = p_left[i] - p_right[i] - l_borrow;
        if(l_diff != p_left[i])
        {
            l_borrow = (l_diff > p_left[i]);
        }
        p_result[i] = l_diff;
    }
    return l_borrow;
}
static void muladd(uECC_word_t a, uECC_word_t b, uECC_word_t *r0, uECC_word_t *r1, uECC_word_t *r2)
{
    uECC_dword_t p = (uECC_dword_t)a * b;
    uECC_dword_t r01 = ((uECC_dword_t)(*r1) << 32) | *r0;
    r01 += p;
    *r2 += (r01 < p);
    *r1 = r01 >> 32;
    *r0 = (uECC_word_t)r01;
}
static void vli_mult(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t r0 = 0;
    uECC_word_t r1 = 0;
    uECC_word_t r2 = 0;
    wordcount_t i, k;
    for(k = 0; k < 8; ++k)
    {
        for(i = 0; i <= k; ++i)
        {
            muladd(p_left[i], p_right[k-i], &r0, &r1, &r2);
        }
        p_result[k] = r0;
        r0 = r1;
        r1 = r2;
        r2 = 0;
    }
    for(k = 8; k < 8*2 - 1; ++k)
    {
        for(i = (k + 1) - 8; i<8; ++i)
        {
            muladd(p_left[i], p_right[k-i], &r0, &r1, &r2);
        }
        p_result[k] = r0;
        r0 = r1;
        r1 = r2;
        r2 = 0;
    }
    p_result[8*2 - 1] = r0;
}
static void mul2add(uECC_word_t a, uECC_word_t b, uECC_word_t *r0, uECC_word_t *r1, uECC_word_t *r2)
{
    uECC_dword_t p = (uECC_dword_t)a * b;
    uECC_dword_t r01 = ((uECC_dword_t)(*r1) << 32) | *r0;
    *r2 += (p >> (32 * 2 - 1));
    p *= 2;
    r01 += p;
    *r2 += (r01 < p);
    *r1 = r01 >> 32;
    *r0 = (uECC_word_t)r01;
}
static void vli_square(uECC_word_t *p_result, uECC_word_t *p_left)
{
    uECC_word_t r0 = 0;
    uECC_word_t r1 = 0;
    uECC_word_t r2 = 0;
    wordcount_t i, k;
    for(k = 0; k < 8*2 - 1; ++k)
    {
        uECC_word_t l_min = (k < 8 ? 0 : (k + 1) - 8);
        for(i = l_min; i<=k && i<=k-i; ++i)
        {
            if(i < k-i)
            {
                mul2add(p_left[i], p_left[k-i], &r0, &r1, &r2);
            }
            else
            {
                muladd(p_left[i], p_left[k-i], &r0, &r1, &r2);
            }
        }
        p_result[k] = r0;
        r0 = r1;
        r1 = r2;
        r2 = 0;
    }
    p_result[8*2 - 1] = r0;
}
static void vli_modAdd(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right, uECC_word_t *p_mod)
{
    uECC_word_t l_carry = vli_add(p_result, p_left, p_right);
    if(l_carry || vli_cmp(p_result, p_mod) >= 0)
    {
        vli_sub(p_result, p_result, p_mod);
    }
}
static void vli_modSub(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right, uECC_word_t *p_mod)
{
    uECC_word_t l_borrow = vli_sub(p_result, p_left, p_right);
    if(l_borrow)
    {
        vli_add(p_result, p_result, p_mod);
    }
}
static void omega_mult(uECC_word_t * p_result, uECC_word_t * p_right);
static void vli_mmod_fast(uECC_word_t * p_result, uECC_word_t * p_product)
{
    uECC_word_t l_tmp[2*8];
    uECC_word_t l_carry;
    vli_clear(l_tmp);
    vli_clear(l_tmp + 8);
    omega_mult(l_tmp, p_product + 8);
    l_carry = vli_add(p_result, p_product, l_tmp);
    vli_clear(p_product);
    omega_mult(p_product, l_tmp + 8);
    l_carry += vli_add(p_result, p_result, p_product);
    while(l_carry > 0)
    {
        --l_carry;
        vli_sub(p_result, p_result, curve_p);
    }
    if(vli_cmp(p_result, curve_p) > 0)
    {
        vli_sub(p_result, p_result, curve_p);
    }
}
static void omega_mult(uint32_t * p_result, uint32_t * p_right)
{
    uint32_t l_carry = 0;
    wordcount_t k;
    for(k = 0; k < 8; ++k)
    {
        uint64_t p = (uint64_t)0x3D1 * p_right[k] + l_carry;
        p_result[k] = (p & 0xffffffff);
        l_carry = p >> 32;
    }
    p_result[8] = l_carry;
    p_result[1 + 8] = vli_add(p_result + 1, p_result + 1, p_right);
}
static void vli_modMult_fast(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_product[2 * 8];
    vli_mult(l_product, p_left, p_right);
    vli_mmod_fast(p_result, l_product);
}
static void vli_modSquare_fast(uECC_word_t *p_result, uECC_word_t *p_left)
{
    uECC_word_t l_product[2 * 8];
    vli_square(l_product, p_left);
    vli_mmod_fast(p_result, l_product);
}
static void vli_modInv(uECC_word_t *p_result, uECC_word_t *p_input, uECC_word_t *p_mod)
{
    uECC_word_t a[8], b[8], u[8], v[8];
    uECC_word_t l_carry;
    cmpresult_t l_cmpResult;
    if(vli_isZero(p_input))
    {
        vli_clear(p_result);
        return;
    }
    vli_set(a, p_input);
    vli_set(b, p_mod);
    vli_clear(u);
    u[0] = 1;
    vli_clear(v);
    while((l_cmpResult = vli_cmp(a, b)) != 0)
    {
        l_carry = 0;
        if((!(a[0] & 1)))
        {
            vli_rshift1(a);
            if(!(!(u[0] & 1)))
            {
                l_carry = vli_add(u, u, p_mod);
            }
            vli_rshift1(u);
            if(l_carry)
            {
                u[8 -1] |= 0x80000000;
            }
        }
        else if((!(b[0] & 1)))
        {
            vli_rshift1(b);
            if(!(!(v[0] & 1)))
            {
                l_carry = vli_add(v, v, p_mod);
            }
            vli_rshift1(v);
            if(l_carry)
            {
                v[8 -1] |= 0x80000000;
            }
        }
        else if(l_cmpResult > 0)
        {
            vli_sub(a, a, b);
            vli_rshift1(a);
            if(vli_cmp(u, v) < 0)
            {
                vli_add(u, u, p_mod);
            }
            vli_sub(u, u, v);
            if(!(!(u[0] & 1)))
            {
                l_carry = vli_add(u, u, p_mod);
            }
            vli_rshift1(u);
            if(l_carry)
            {
                u[8 -1] |= 0x80000000;
            }
        }
        else
        {
            vli_sub(b, b, a);
            vli_rshift1(b);
            if(vli_cmp(v, u) < 0)
            {
                vli_add(v, v, p_mod);
            }
            vli_sub(v, v, u);
            if(!(!(v[0] & 1)))
            {
                l_carry = vli_add(v, v, p_mod);
            }
            vli_rshift1(v);
            if(l_carry)
            {
                v[8 -1] |= 0x80000000;
            }
        }
    }
    vli_set(p_result, u);
}
static cmpresult_t EccPoint_isZero(EccPoint *p_point)
{
    return (vli_isZero(p_point->x) && vli_isZero(p_point->y));
}
static void EccPoint_double_jacobian(uECC_word_t * X1, uECC_word_t * Y1, uECC_word_t * Z1)
{
    uECC_word_t t4[8];
    uECC_word_t t5[8];
    if(vli_isZero(Z1))
    {
        return;
    }
    vli_modSquare_fast(t5, Y1);
    vli_modMult_fast(t4, X1, t5);
    vli_modSquare_fast(X1, X1);
    vli_modSquare_fast(t5, t5);
    vli_modMult_fast(Z1, Y1, Z1);
    vli_modAdd(Y1, X1, X1, curve_p);
    vli_modAdd(Y1, Y1, X1, curve_p);
    if(vli_testBit(Y1, 0))
    {
        uECC_word_t l_carry = vli_add(Y1, Y1, curve_p);
        vli_rshift1(Y1);
        Y1[8 -1] |= l_carry << (32 - 1);
    }
    else
    {
        vli_rshift1(Y1);
    }
    vli_modSquare_fast(X1, Y1);
    vli_modSub(X1, X1, t4, curve_p);
    vli_modSub(X1, X1, t4, curve_p);
    vli_modSub(t4, t4, X1, curve_p);
    vli_modMult_fast(Y1, Y1, t4);
    vli_modSub(Y1, Y1, t5, curve_p);
}
static void apply_z(uECC_word_t * X1, uECC_word_t * Y1, uECC_word_t * Z)
{
    uECC_word_t t1[8];
    vli_modSquare_fast(t1, Z);
    vli_modMult_fast(X1, X1, t1);
    vli_modMult_fast(t1, t1, Z);
    vli_modMult_fast(Y1, Y1, t1);
}
static void XYcZ_initial_double(uECC_word_t * X1, uECC_word_t * Y1,
    uECC_word_t * X2, uECC_word_t * Y2, const uECC_word_t * p_initialZ)
{
    uECC_word_t z[8];
    vli_set(X2, X1);
    vli_set(Y2, Y1);
    vli_clear(z);
    z[0] = 1;
    if(p_initialZ)
    {
        vli_set(z, p_initialZ);
    }
    apply_z(X1, Y1, z);
    EccPoint_double_jacobian(X1, Y1, z);
    apply_z(X2, Y2, z);
}
static void XYcZ_add(uECC_word_t * X1, uECC_word_t * Y1, uECC_word_t * X2, uECC_word_t * Y2)
{
    uECC_word_t t5[8];
    vli_modSub((t5), (X2), (X1), curve_p);
    vli_modSquare_fast(t5, t5);
    vli_modMult_fast(X1, X1, t5);
    vli_modMult_fast(X2, X2, t5);
    vli_modSub((Y2), (Y2), (Y1), curve_p);
    vli_modSquare_fast(t5, Y2);
    vli_modSub((t5), (t5), (X1), curve_p);
    vli_modSub((t5), (t5), (X2), curve_p);
    vli_modSub((X2), (X2), (X1), curve_p);
    vli_modMult_fast(Y1, Y1, X2);
    vli_modSub((X2), (X1), (t5), curve_p);
    vli_modMult_fast(Y2, Y2, X2);
    vli_modSub((Y2), (Y2), (Y1), curve_p);
    vli_set(X2, t5);
}
static void XYcZ_addC(uECC_word_t * X1, uECC_word_t * Y1, uECC_word_t * X2, uECC_word_t * Y2)
{
    uECC_word_t t5[8];
    uECC_word_t t6[8];
    uECC_word_t t7[8];
    vli_modSub((t5), (X2), (X1), curve_p);
    vli_modSquare_fast(t5, t5);
    vli_modMult_fast(X1, X1, t5);
    vli_modMult_fast(X2, X2, t5);
    vli_modAdd(t5, Y2, Y1, curve_p);
    vli_modSub((Y2), (Y2), (Y1), curve_p);
    vli_modSub((t6), (X2), (X1), curve_p);
    vli_modMult_fast(Y1, Y1, t6);
    vli_modAdd(t6, X1, X2, curve_p);
    vli_modSquare_fast(X2, Y2);
    vli_modSub((X2), (X2), (t6), curve_p);
    vli_modSub((t7), (X1), (X2), curve_p);
    vli_modMult_fast(Y2, Y2, t7);
    vli_modSub((Y2), (Y2), (Y1), curve_p);
    vli_modSquare_fast(t7, t5);
    vli_modSub((t7), (t7), (t6), curve_p);
    vli_modSub((t6), (t7), (X1), curve_p);
    vli_modMult_fast(t6, t6, t5);
    vli_modSub((Y1), (t6), (Y1), curve_p);
    vli_set(X1, t7);
}
static void EccPoint_mult(EccPoint * p_result, EccPoint * p_point,
    const uECC_word_t * p_scalar, const uECC_word_t * p_initialZ, bitcount_t p_numBits)
{
    uECC_word_t Rx[2][8];
    uECC_word_t Ry[2][8];
    uECC_word_t z[8];
    bitcount_t i;
    uECC_word_t nb;
    vli_set(Rx[1], p_point->x);
    vli_set(Ry[1], p_point->y);
    XYcZ_initial_double(Rx[1], Ry[1], Rx[0], Ry[0], p_initialZ);
    for(i = p_numBits - 2; i > 0; --i)
    {
        nb = !vli_testBit(p_scalar, i);
        XYcZ_addC(Rx[1-nb], Ry[1-nb], Rx[nb], Ry[nb]);
        XYcZ_add(Rx[nb], Ry[nb], Rx[1-nb], Ry[1-nb]);
    }
    nb = !vli_testBit(p_scalar, 0);
    XYcZ_addC(Rx[1-nb], Ry[1-nb], Rx[nb], Ry[nb]);
    vli_modSub((z), (Rx[1]), (Rx[0]), curve_p);
    vli_modMult_fast(z, z, Ry[1-nb]);
    vli_modMult_fast(z, z, p_point->x);
    vli_modInv(z, z, curve_p);
    vli_modMult_fast(z, z, p_point->y);
    vli_modMult_fast(z, z, Rx[1-nb]);
    XYcZ_add(Rx[nb], Ry[nb], Rx[1-nb], Ry[1-nb]);
    apply_z(Rx[0], Ry[0], z);
    vli_set(p_result->x, Rx[0]);
    vli_set(p_result->y, Ry[0]);
}
static void mod_sqrt(uECC_word_t *a)
{
    bitcount_t i;
    uECC_word_t p1[8] = {1};
    uECC_word_t l_result[8] = {1};
    vli_add(p1, curve_p, p1);
    for(i = vli_numBits(p1, 8) - 1; i > 1; --i)
    {
        vli_modSquare_fast(l_result, l_result);
        if(vli_testBit(p1, i))
        {
            vli_modMult_fast(l_result, l_result, a);
        }
    }
    vli_set(a, l_result);
}
static void vli_nativeToBytes(uint8_t *p_bytes, const uint32_t *p_native)
{
    unsigned i;
    for(i=0; i<8; ++i)
    {
        uint8_t *p_digit = p_bytes + 4 * (8 - 1 - i);
        p_digit[0] = p_native[i] >> 24;
        p_digit[1] = p_native[i] >> 16;
        p_digit[2] = p_native[i] >> 8;
        p_digit[3] = p_native[i];
    }
}
static void vli_bytesToNative(uint32_t *p_native, const uint8_t *p_bytes)
{
    unsigned i;
    for(i=0; i<8; ++i)
    {
        const uint8_t *p_digit = p_bytes + 4 * (8 - 1 - i);
        p_native[i] = ((uint32_t)p_digit[0] << 24) | ((uint32_t)p_digit[1] << 16) | ((uint32_t)p_digit[2] << 8) | (uint32_t)p_digit[3];
    }
}
int uECC_make_key(uint8_t p_publicKey[32*2], uint8_t p_privateKey[32])
{
    EccPoint l_public;
    uECC_word_t l_private[8];
    uECC_word_t l_tries = 0;
    do
    {
    repeat:
        if(!g_rng((uint8_t *)l_private, sizeof(l_private)) || (l_tries++ >= 16))
        {
            return 0;
        }
        if(vli_isZero(l_private))
        {
            goto repeat;
        }
        if(vli_cmp(curve_n, l_private) != 1)
        {
            goto repeat;
        }
        EccPoint_mult(&l_public, &curve_G, l_private, 0, vli_numBits(l_private, 8));
    } while(EccPoint_isZero(&l_public));
    vli_nativeToBytes(p_privateKey, l_private);
    vli_nativeToBytes(p_publicKey, l_public.x);
    vli_nativeToBytes(p_publicKey + 32, l_public.y);
    return 1;
}
int uECC_shared_secret(const uint8_t p_publicKey[32*2], const uint8_t p_privateKey[32], uint8_t p_secret[32])
{
    EccPoint l_public;
    uECC_word_t l_private[8];
    uECC_word_t l_random[8];
    g_rng((uint8_t *)l_random, sizeof(l_random));
    vli_bytesToNative(l_private, p_privateKey);
    vli_bytesToNative(l_public.x, p_publicKey);
    vli_bytesToNative(l_public.y, p_publicKey + 32);
    EccPoint l_product;
    EccPoint_mult(&l_product, &l_public, l_private, (vli_isZero(l_random) ? 0: l_random), vli_numBits(l_private, 8));
    vli_nativeToBytes(p_secret, l_product.x);
    return !EccPoint_isZero(&l_product);
}
void uECC_compress(const uint8_t p_publicKey[32*2], uint8_t p_compressed[32 +1])
{
    wordcount_t i;
    for(i=0; i<32; ++i)
    {
        p_compressed[i+1] = p_publicKey[i];
    }
    p_compressed[0] = 2 + (p_publicKey[32 * 2 - 1] & 0x01);
}
void uECC_decompress(const uint8_t p_compressed[32 +1], uint8_t p_publicKey[32*2])
{
    EccPoint l_point;
    vli_bytesToNative(l_point.x, p_compressed + 1);
    vli_modSquare_fast(l_point.y, l_point.x);
    vli_modMult_fast(l_point.y, l_point.y, l_point.x);
    vli_modAdd(l_point.y, l_point.y, curve_b, curve_p);
    mod_sqrt(l_point.y);
    if((l_point.y[0] & 0x01) != (p_compressed[0] & 0x01))
    {
        vli_sub(l_point.y, curve_p, l_point.y);
    }
    vli_nativeToBytes(p_publicKey, l_point.x);
    vli_nativeToBytes(p_publicKey + 32, l_point.y);
}
static void vli2_rshift1(uECC_word_t *p_vli)
{
    vli_rshift1(p_vli);
    p_vli[8 -1] |= p_vli[8] << (32 - 1);
    vli_rshift1(p_vli + 8);
}
static uECC_word_t vli2_sub(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_borrow = 0;
    wordcount_t i;
    for(i=0; i<8*2; ++i)
    {
        uECC_word_t l_diff = p_left[i] - p_right[i] - l_borrow;
        if(l_diff != p_left[i])
        {
            l_borrow = (l_diff > p_left[i]);
        }
        p_result[i] = l_diff;
    }
    return l_borrow;
}
static void vli_modMult_n(uECC_word_t *p_result, uECC_word_t *p_left, uECC_word_t *p_right)
{
    uECC_word_t l_product[2 * 8];
    uECC_word_t l_modMultiple[2 * 8];
    uECC_word_t l_tmp[2 * 8];
    uECC_word_t *v[2] = {l_tmp, l_product};
    vli_mult(l_product, p_left, p_right);
    vli_set(l_modMultiple + 8, curve_n);
    vli_clear(l_modMultiple);
    bitcount_t i;
    uECC_word_t l_index = 1;
    for(i=0; i<=32 * 8; ++i)
    {
        uECC_word_t l_borrow = vli2_sub(v[1-l_index], v[l_index], l_modMultiple);
        l_index = !(l_index ^ l_borrow);
        vli2_rshift1(l_modMultiple);
    }
    vli_set(p_result, v[l_index]);
}
int uECC_sign(const uint8_t p_privateKey[32], const uint8_t p_hash[32], uint8_t p_signature[32*2])
{
    uECC_word_t k[8];
    uECC_word_t l_tmp[8];
    uECC_word_t s[8];
    uECC_word_t *k2[2] = {l_tmp, s};
    EccPoint p;
    uECC_word_t l_tries = 0;
    do
    {
    repeat:
        if(!g_rng((uint8_t *)k, sizeof(k)) || (l_tries++ >= 16))
        {
            return 0;
        }
        if(vli_isZero(k))
        {
            goto repeat;
        }
        if(vli_cmp(curve_n, k) != 1)
        {
            goto repeat;
        }
        uECC_word_t l_carry = vli_add(l_tmp, k, curve_n);
        vli_add(s, l_tmp, curve_n);
        EccPoint_mult(&p, &curve_G, k2[!l_carry], 0, (32 * 8) + 1);
        if(vli_cmp(curve_n, p.x) != 1)
        {
            vli_sub(p.x, p.x, curve_n);
        }
    } while(vli_isZero(p.x));
    l_tries = 0;
    do
    {
        if(!g_rng((uint8_t *)l_tmp, sizeof(l_tmp)) || (l_tries++ >= 16))
        {
            return 0;
        }
    } while(vli_isZero(l_tmp));
    vli_modMult_n(k, k, l_tmp);
    vli_modInv(k, k, curve_n);
    vli_modMult_n(k, k, l_tmp);
    vli_nativeToBytes(p_signature, p.x);
    l_tmp[8 -1] = 0;
    vli_bytesToNative(l_tmp, p_privateKey);
    s[8 -1] = 0;
    vli_set(s, p.x);
    vli_modMult_n(s, l_tmp, s);
    vli_bytesToNative(l_tmp, p_hash);
    vli_modAdd(s, l_tmp, s, curve_n);
    vli_modMult_n(s, s, k);
    vli_nativeToBytes(p_signature + 32, s);
    return 1;
}
static bitcount_t smax(bitcount_t a, bitcount_t b)
{
    return (a > b ? a : b);
}
int uECC_verify(const uint8_t p_publicKey[32*2], const uint8_t p_hash[32], const uint8_t p_signature[32*2])
{
    uECC_word_t u1[8], u2[8];
    uECC_word_t z[8];
    EccPoint l_public, l_sum;
    uECC_word_t rx[8];
    uECC_word_t ry[8];
    uECC_word_t tx[8];
    uECC_word_t ty[8];
    uECC_word_t tz[8];
    uECC_word_t r[8], s[8];
    r[8 -1] = 0;
    s[8 -1] = 0;
    vli_bytesToNative(l_public.x, p_publicKey);
    vli_bytesToNative(l_public.y, p_publicKey + 32);
    vli_bytesToNative(r, p_signature);
    vli_bytesToNative(s, p_signature + 32);
    if(vli_isZero(r) || vli_isZero(s))
    {
        return 0;
    }
    if(vli_cmp(curve_n, r) != 1 || vli_cmp(curve_n, s) != 1)
    {
        return 0;
    }
    vli_modInv(z, s, curve_n);
    u1[8 -1] = 0;
    vli_bytesToNative(u1, p_hash);
    vli_modMult_n(u1, u1, z);
    vli_modMult_n(u2, r, z);
    vli_set(l_sum.x, l_public.x);
    vli_set(l_sum.y, l_public.y);
    vli_set(tx, curve_G.x);
    vli_set(ty, curve_G.y);
    vli_modSub((z), (l_sum.x), (tx), curve_p);
    XYcZ_add(tx, ty, l_sum.x, l_sum.y);
    vli_modInv(z, z, curve_p);
    apply_z(l_sum.x, l_sum.y, z);
    EccPoint *l_points[4] = {0, &curve_G, &l_public, &l_sum};
    bitcount_t l_numBits = smax(vli_numBits(u1, 8), vli_numBits(u2, 8));
    EccPoint *l_point = l_points[(!!vli_testBit(u1, l_numBits-1)) | ((!!vli_testBit(u2, l_numBits-1)) << 1)];
    vli_set(rx, l_point->x);
    vli_set(ry, l_point->y);
    vli_clear(z);
    z[0] = 1;
    bitcount_t i;
    for(i = l_numBits - 2; i >= 0; --i)
    {
        EccPoint_double_jacobian(rx, ry, z);
        uECC_word_t l_index = (!!vli_testBit(u1, i)) | ((!!vli_testBit(u2, i)) << 1);
        l_point = l_points[l_index];
        if(l_point)
        {
            vli_set(tx, l_point->x);
            vli_set(ty, l_point->y);
            apply_z(tx, ty, z);
            vli_modSub((tz), (rx), (tx), curve_p);
            XYcZ_add(tx, ty, rx, ry);
            vli_modMult_fast(z, z, tz);
        }
    }
    vli_modInv(z, z, curve_p);
    apply_z(rx, ry, z);
    if(vli_cmp(curve_n, rx) != 1)
    {
        vli_sub(rx, rx, curve_n);
    }
    return vli_equal(rx, r);
}
int uECC_bytes()
{
    return 32;
}
int uECC_curve()
{
    return 4;
}
